/*
 * This file is part of Firestorm NIDS
 * Copyright (c) 2002,2003,2004 Gianni Tedesco <gianni@scaramanga.co.uk>
 * This program is released under the terms of the GNU GPL version 2
*/

#include <firestorm.h>

#include <sys/socket.h>
#include <netinet/in.h>

#include <args.h>
#include <plugin.h>
#include <alert.h>
#include <signature.h>
#include <decode.h>
#include <pkt/arp.h>


/* Token bucket for arp alerts */
static struct tokenbucket arp_tb={
	.cost = (RATE_SEC / 1), /* 1 alert per second */
	.burst = (RATE_SEC / 1) * 10, /* a burst of 10 alerts per second */
};

static struct generator arp_gen=init_generator("arp", &arp_tb);

static struct alert alert_trunc={
	.alert="Truncated arp packet",
	.sid=5, .rev=0,
	.priority=5,
	.gen=&arp_gen,
	.order = ALERT_BUILTIN,
};


void arp_decode(struct packet *);
int arp_print(struct packet *p, int l, char *buf, size_t buflen);
void rarp_decode(struct packet *);

struct proto rarp_p=init_proto("rarp", rarp_decode, NULL);
struct proto_req rarp_r[]={
	proto_request("ethernet", __constant_htons(0x0835)),
	proto_request("sll", __constant_htons(0x0835)),
	proto_request("linux", __constant_htons(0x0835)),
	null_request()
};

struct proto arp_p=init_proto("arp", arp_decode, arp_print);
struct proto_req arp_r[]={
	/* IP ARP */
	proto_request("ethernet", __constant_htons(0x0806)),
	proto_request("sll", __constant_htons(0x0806)),
	proto_request("linux", __constant_htons(0x0806)),

	/* AppleTalk ARP */
	proto_request("ethernet", __constant_htons(0x80f3)),
	proto_request("sll", __constant_htons(0x80f3)),
	proto_request("linux", __constant_htons(0x80f3)),

	null_request()
};

char *opnames[] = { NULL,
	/* RFC 826 */
	"hw req","hw reply",
	/* RFC 903 */
	"Proto req","Proto reply",
	/* RFC 1931 */
	"DRARP req", "DRARP reply", "DRARP error",
	/* RFC 1293 */
	"InARP req","InARP reply",
	/* RFC 1577 */
	"ARP nak",
	/* Multicast Address Resolution Server */
	"MARS req", "MARS multi", "MARS mserv", "MARS join", "MARS leave",
	"MARS nak", "MARS unserv", "MARS sjoin", "MARS sleave",
	"MARS grouplist request", "MARS grouplist reply", "MARS redirect map",
	"MAPOS unarp"
};

char *hwtypes[] = { NULL,
	"Ethernet",
	"Experimental ethernet",
	"AX.25",
	"ProNET token ring",
	"Chaos",
	"IEEE 802",
	"ARCNET",
	"Hyperchannel",
	"Lanstar",
	"Autonet short address",
	"LocalTalk",
	"LocalNet",
	"Ultra link",
	"SMDS",
	"Frame Relay",
	"ATM",
	"HDLC",
	"Fibre channel",
	"ATM",
	"Serial line",
	"ATM",
	"MIL-STD-188-220",
	"Metricom",
	"IEEE 1394.1995",
	"MAPOS",
	"Twinaxial",
	"EUI-64",
	"HIPARP",
	"ISO7816-3",
	"ARPSec",
	"IPsec tunnel",
	"Infiniband"
};

void arp_decode(struct packet *p)
{
	struct layer *l = &p->layer[p->llen];
	void *payload = l->h.raw + sizeof(struct pkt_arphdr);
	u_int8_t hln = l->h.arp->hln;
	u_int8_t pln = l->h.arp->pln;

	/* Check length of packet */
	if (payload + (hln*2) + (pln*2) <= p->end) {
		l->flags |= FLAG_ARP_TARGET | FLAG_ARP_SENDER;
	} else if (payload + hln + pln <= p->end) {
		/* Linux seems to process truncated arp packets with sender
		 * addresses intact */
		l->flags |= FLAG_ARP_SENDER;
	} else {
		/* Truncated arp packet */
		/* With Ethernet, this is actually very unlikely due to
		 * Ethernet frame padding to minimum size of FIXME: 60 bytes.
		 * We just end up using bytes from this padding as
		 * sender/target addresses if the arp packet is truncated
		 * somewhere after the header.   It looks like Linux does this
		 * too. */
		alert_tag(p, &alert_trunc, -1);
		dispatch(p);
		return;
	}

	p->llen++;

	dispatch(p);
}

/* print a variable length hardware address in ascii (00:11:22...) */
static int print_haddr(char *s, int slen, u_int8_t *b, u_int8_t hln)
{
	int i;

	if ((hln*3)+1 > slen)
		return 0;

	for(i = 1 ; i <= hln; i++, b++)
		s+=snprintf(s, 4, "%02x%c", *b & 0xff, i == hln ? '\0' : ':');

	return 1;
}

/* print a variable length protocol address in ascii (10.9.8...) */
static int print_paddr(char *s, int slen, u_int8_t *b, u_int8_t pln)
{
	int i;

	if ((pln*4)+1 > slen)
		return 0;

	for(i = 1 ; i <= pln; i++, b++)
		s+=snprintf(s, 5, "%u%c", *b & 0xff, i == pln ? '\0' : '.');

	return 1;
}

int arp_print(struct packet *p, int l, char *buf, size_t buflen)
{
	struct pkt_arphdr *a = p->layer[l].h.arp;
	u_int16_t op = ntohs(a->op);
	u_int16_t hrd = ntohs(a->hrd);
	u_int8_t hln = a->hln;
	u_int8_t pln = a->pln;
	u_int8_t *payload = p->layer[l].h.raw + sizeof(struct pkt_arphdr);
	char shwtype[32];
	char sop[32];
	char ssh[24]="n/a", ssp[24]="n/a";
	char tsh[24]="n/a", tsp[24]="n/a";

	/* print hardware type */
	if ( hrd < (sizeof(hwtypes)/sizeof(*hwtypes)) && hwtypes[hrd] ) {
		snprintf(shwtype, sizeof(shwtype), hwtypes[hrd]);
	}else{
		snprintf(shwtype, sizeof(shwtype),
			"Unknown hw type (0x%.4x)", hrd);
	}

	/* print op code */
	if ( op < (sizeof(opnames)/sizeof(*opnames)) && opnames[op] ) {
		snprintf(sop, sizeof(sop), opnames[op]); /* not user input */
	}else{
		snprintf(sop, sizeof(sop),
			"Unknown opcode (0x%.4x)", op);
	}

	if (p->layer[l].flags & FLAG_ARP_SENDER) {
		print_haddr(ssh, sizeof(ssh), payload, hln);
		print_paddr(ssp, sizeof(ssp), payload + hln, pln);
	}

	if (p->layer[l].flags & FLAG_ARP_TARGET) {
		print_haddr(tsh, sizeof(tsh), payload + hln + pln, hln);
		print_paddr(tsp, sizeof(tsp), payload + hln + pln + hln, pln);
	}

	switch (op) {
	/* Request */
	case 1:
		return snprintf(buf, buflen, "%s %s: who has %s? tell %s (%s)",
				shwtype, sop, tsp, ssp, ssh);
	/* Reply */
	case 2:
		return snprintf(buf, buflen, "%s %s: %s is at %s",
				shwtype, sop, ssp, ssh);
	default:
		return snprintf(buf, buflen, "%s %s %s/%s > %s/%s",
				shwtype, sop, ssh, ssp, tsh, tsp);
	}

}

void rarp_decode(struct packet *p)
{
	struct layer *l=&p->layer[p->llen];

	if ( l->h.raw+sizeof(struct pkt_arphdr) > p->end ) {
		dispatch(p);
		return;
	}

	p->llen++;

	dispatch(p);
}

struct proto_table firestorm_protocols[]={
	{.proto = &arp_p, .preq = arp_r},
	{.proto = &rarp_p, .preq = rarp_r},
	{.proto = NULL},
};

struct plugin_desc firestorm_plugin_desc = {
	.size		= sizeof(firestorm_plugin_desc),
	.name		= "decode.arp",
	.desc		= "ARP/RARP",
	.author_name	= "Gianni Tedesco",
	.author_email	= "gianni@scaramanga.co.uk",
	.ver_major	= 2,
	.ver_minor	= 1,
};
