/*
 * This file is part of Firestorm NIDS
 * Copyright (c) 2002,2003,2004 Gianni Tedesco <gianni@scaramanga.co.uk>
 * This program is released under the terms of the GNU GPL version 2
*/
#include <firestorm.h>
#include <dotscara/sys.h>

#include <sys/socket.h>
#include <netinet/in.h>

#include <args.h>
#include <plugin.h>
#include <alert.h>
#include <signature.h>
#include <decode.h>

struct linux_sll {
	unsigned short	sll_family;
	unsigned short	sll_protocol;
	int		sll_ifindex;
	unsigned short	sll_hatype;
	unsigned char	sll_pkttype;
	unsigned char	sll_halen;
	unsigned char	sll_addr[8];
};

static const char *ptype[]={
	"unicast",
	"broadcast",
	"multicast",
	"promisc",
	"outgoing",
	"loopback",
	"fastroute"
};

static int sll_ha_name(unsigned short hatype, char *buf, size_t buflen)
{
	if ( hatype == 1)
		return snprintf(buf, buflen, "ethernet");
	if ( hatype == 772)
		return snprintf(buf, buflen, "loopback");

	return snprintf(buf, 16, "type 0x%.4x", hatype);
}

static int sll_print(struct packet *p, int l, char *buf, size_t buflen)
{
	struct pkt_sllhdr *sll = p->layer[l].h.sll;
	const char *type;
	u_int16_t pt;

	pt = ntohs(sll->sll_pkttype);
	type = ( pt < 7) ? ptype[pt] : "unknown";

	return snprintf(buf, buflen,"%s proto=0x%.4x",
		type, ntohs(sll->sll_protocol));
}

static int lsll_print(struct packet *p, int l, char *buf, size_t buflen)
{
	struct linux_sll *sll = p->layer[l].h.raw;
	char hwaddr[sizeof(sll->sll_addr)*3];
	char hdr[16];
	const char *type;
	unsigned int i, halen;

	type = (sll->sll_pkttype < 7) ?
		ptype[sll->sll_pkttype] :
		"unknown";

	halen = ( sll->sll_halen > sizeof(sll->sll_addr) ) ?
		sizeof(sll->sll_addr) :
		sll->sll_halen;

	sll_ha_name(pkt16(p, sll->sll_hatype), hdr, sizeof(hdr));

	if ( halen == 0 ) {
		hwaddr[0] = '\0';
	}else{
		for (i = 0; i < halen; i++)
			sprintf(hwaddr + i*3, "%.2x:", sll->sll_addr[i]);
		hwaddr[(i*3)-1] = '\0';
	}

	return snprintf(buf, buflen,
		"if%i:%s - %s %s",
		pkt32(p, sll->sll_ifindex),
		type,
		hdr,
		hwaddr);
}

static int null_print(struct packet *p, int l, char *buf, size_t buflen)
{
	u_int32_t *type = p->layer[l].h.raw;
	return snprintf(buf, buflen, "type=0x%x", pkt32(p, *type));
}

static void lsll_decode(struct packet *p)
{
	struct proto *child;
	struct layer *l=&p->layer[p->llen];
	struct linux_sll *sll=(struct linux_sll *)l->h.raw;

	if ( (p->layer[p->llen+1].h.raw=
		l->h.raw+48)
		> p->end ) goto finish;

	p->llen++;

	if ( p->llen >= PKT_LAYERS ) {
		dispatch(p);
		return;
	}

	child = decode_next(l->proto, sll->sll_protocol);
	if ( child ) {
		p->layer[p->llen].flags = 0;
		p->layer[p->llen].session = NULL;
		p->layer[p->llen].proto = child;
		child->decode(p);
		return;
	}

finish:
	if ( p->layer[p->llen].h.raw < p->end )
		p->layer[p->llen].proto = NULL;

	dispatch(p);
}

static void sll_decode(struct packet *p)
{
	struct proto *child;
	struct layer *l=&p->layer[p->llen];

	if ( (p->layer[p->llen+1].h.raw=
		l->h.raw+sizeof(struct pkt_sllhdr))
		> p->end ) goto finish;

	/*
	 * If this is the first layer, then perhaps
	 * we should set the packet type
	*/
	if ( p->llen==0 && !(p->flags&FP_PKTTYPE) ) {
		p->flags&=~FPMASK_PKTTYPE;
		p->flags|=FP_PKTTYPE;
		switch(htons(l->h.sll->sll_pkttype))
		{
			case 0:
				p->flags|=FP_HOST;
				break;
			case 1:
				p->flags|=FP_BROADCAST;
				break;
			case 2:
				p->flags|=FP_MULTICAST;
				break;
			case 4:
				p->flags|=FP_OUTGOING;
				break;
			case 3:
			default:
				p->flags|=FP_PROMISC;
				break;
		}
	}

	p->llen++;

	if ( p->llen >= PKT_LAYERS ) {
		dispatch(p);
		return;
	}

	child = decode_next(l->proto, l->h.sll->sll_protocol);
	if ( child ) {
		p->layer[p->llen].flags = 0;
		p->layer[p->llen].session = NULL;
		p->layer[p->llen].proto = child;
		child->decode(p);
		return;
	}

finish:
	if ( p->layer[p->llen].h.raw < p->end )
		p->layer[p->llen].proto = NULL;

	dispatch(p);
}

static void null_decode(struct packet *p)
{
	struct proto *child;
	struct layer *l=&p->layer[p->llen];

	if ( (p->layer[p->llen+1].h.raw = l->h.raw + 4) > p->end )
		goto finish;

	p->llen++;

	if ( p->llen >= PKT_LAYERS ) {
		dispatch(p);
		return;
	}

	child = decode_next(l->proto, pkt32(p, *(u_int32_t *)l->h.raw));
	if ( child ) {
		p->layer[p->llen].flags = 0;
		p->layer[p->llen].session = NULL;
		p->layer[p->llen].proto = child;
		child->decode(p);
		return;
	}

finish:
	if ( p->layer[p->llen].h.raw < p->end )
		p->layer[p->llen].proto = NULL;

	dispatch(p);
}

static struct proto sll_p=init_proto("sll", sll_decode, sll_print);
static struct proto lsll_p=init_proto("linux", lsll_decode, lsll_print);
static struct proto null_p=init_proto("null", null_decode, null_print);
static struct proto_req lsll_r[]={
	null_request()
};
static struct proto_req sll_r[]={
	proto_request("__pcap_dlt", 0x71),
	null_request()
};
static struct proto_req null_r[]={
	proto_request("__pcap_dlt", 0x0),
	null_request()
};

struct proto_table firestorm_protocols[]={
	{.proto = &sll_p, .preq = sll_r},
	{.proto = &lsll_p, .preq = lsll_r},
	{.proto = &null_p, .preq = null_r},
	{.proto = NULL}
};

struct plugin_desc firestorm_plugin_desc = {
	.size		= sizeof(firestorm_plugin_desc),
	.name		= "decode.sll",
	.desc		= "Linux SLL and others",
	.author_name	= "Gianni Tedesco",
	.author_email	= "gianni@scaramanga.co.uk",
	.ver_major	= 1,
	.ver_minor	= 1,
};
