/*
 * This file is part of Firestorm NIDS
 * Copyright (c) 2002,2003,2004 Gianni Tedesco
 * This program is released under the terms of the GNU GPL version 2
 *
 * This is an argument parsing library designed to work on const strings,
 * it is typesafe and presents a simple API. Supported types are strings,
 * integers (signed and unsigned), double precision floats and boolean.
 *
 * TODO:
 *  o Support validator functions
 *  o Support constraints (lists of possible values)
 *  o Support combos (constraints or user input)
 *  o Return detailed errors
 *  o Support arguments without names to get all non-matching args. This
 *    requires another arg_get_string variant.
*/
#include <firestorm.h>
#include <dotscara/sys.h>
#include <dotscara/fdctl.h>
#include <dotscara/fobuf.h>
#include <dotscara/fibuf.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <limits.h>
#include <math.h>
#include <errno.h>
#include <args.h>

/* first char is blank because arg_q[0] means no quotes */
static char arg_q[]=" \'\"";

/* Dispatch functions */
static int arg_custom(struct arg *, struct fvec *, void *);
static int arg_string(struct arg *, struct fvec *, void *);
static int arg_bytes(struct arg *, struct fvec *, void *);
static int arg_off(struct arg *, struct fvec *, void *);
static int arg_uint(struct arg *, struct fvec *, void *);
static int arg_int(struct arg *, struct fvec *, void *);
static int arg_bool(struct arg *, struct fvec *, void *);
static int arg_double(struct arg *, struct fvec *, void *);

/* Dispatch table */
typedef int(*proc_argdispatch)(struct arg *, struct fvec *, void *);
static proc_argdispatch arg_dfn[ARGTYPE_MAX]={
	[ ARGTYPE_CUSTOM ] arg_custom,
	[ ARGTYPE_STRING ] arg_string,
	[ ARGTYPE_PSTRING ] arg_string,
	[ ARGTYPE_UINT ] arg_uint,
	[ ARGTYPE_PUINT ] arg_uint,
	[ ARGTYPE_INT ] arg_int,
	[ ARGTYPE_PINT ] arg_int,
	[ ARGTYPE_BYTES ] arg_bytes,
	[ ARGTYPE_PBYTES ] arg_bytes,
	[ ARGTYPE_BOOL ] arg_bool,
	[ ARGTYPE_PBOOL ] arg_bool,
	[ ARGTYPE_DOUBLE ] arg_double,
	[ ARGTYPE_PDOUBLE ] arg_double,
	[ ARGTYPE_OFF ] arg_off,
	[ ARGTYPE_POFF ] arg_off,
};

/* Copy and NULL terminate from a read-only vector in to a dynamically
 * allocated buffer converting any escape chars on the way.
 */
char *arg_get_string(struct fvec *v)
{
	const char *src, *end;
	static char *vbuf;
	static size_t vbuflen;
	char *dst;
	int esc;

	if ( v->len == 0 || v->ptr == NULL )
		return NULL;

	if ( v->len + 1 > vbuflen ) {
		char *new;
		new = realloc(vbuf, v->len + 1);
		if ( new == NULL )
			return NULL;

		vbuf = new;
		vbuflen = v->len + 1;
	}

	src = v->ptr;
	end = v->ptr + v->len;
	dst = vbuf;

	/* TODO: Support C-style escape characters */
	for(esc=0; src < end; src++) {
		if ( !esc && *src=='\\' ) {
			esc = 1;
		}else{
			*(dst++) = *src;
			esc = 0;
		}
	}

	*dst = '\0';

	return vbuf;
}

/* Boolean data type */
static int arg_bool(struct arg *a, struct fvec *v, void *user)
{
	unsigned int r;

	if ( v == NULL ) {
		r = !!a->count;
	}else if ( !vstrcmp(v, "true") ||
		!vstrcmp(v, "yes") ||
		!vstrcmp(v, "on") ) {
		r = 1;
	}else if ( !vstrcmp(v, "false") ||
		!vstrcmp(v, "no") ||
		!vstrcmp(v, "off") ) {
		r = 0;
	}else
		return -1;

	if ( a->type == ARGTYPE_PBOOL ) {
		if ( a->val.vp_bool )
			*a->val.vp_bool = r;
	}else{
		a->val.v_bool = r;
		return 1;
	}

	return 1;
}

/* Numeric datatype for representing bytes (eg: 1024k, 10M ) */
static int arg_bytes(struct arg *a, struct fvec *v, void *user)
{
	unsigned long int val;
	char *str;
	char *end;

	if ( v == NULL ) {
		val = 0;
		goto skip;
	}

	str = arg_get_string(v);
	if ( str == NULL )
		return -1;

	val = strtoul(str, &end, 0);
	if ( end == str )
		return -1;

	if ( val == ULONG_MAX && errno == ERANGE )
		return -1;

	switch( *end ) {
	case '\0':
		break;
	case 'k':
	case 'K':
		val <<= 10;
		break;
	case 'm':
	case 'M':
		val <<= 20;
		break;
	case 'g':
	case 'G':
		val <<= 30;
		break;
#if 0
	case 't':
	case 'T':
		val <<= 40;
		break;
	case 'p':
	case 'P':
		val <<= 50;
		break;
	case 'e':
	case 'E':
		val <<= 60;
		break;
	case 'z':
	case 'Z':
		val <<= 70;
		break;
	case 'y':
	case 'Y':
		val <<= 80;
		break;
#endif
	}

	if ( val > UINT_MAX )
		return -1;

skip:
	if ( a->type == ARGTYPE_PBYTES ) {
		if ( a->val.vp_bytes )
			*a->val.vp_bytes = val;
	}else{
		a->val.v_bytes = val;
	}

	return 1;
}

/* Numeric datatype for representing off_t (eg: 1024k, 10M ) */
static int arg_off(struct arg *a, struct fvec *v, void *user)
{
#if _FILE_OFFSET_BITS == 64
	unsigned long long int val;
#else
	unsigned long int val;
#endif
	char *end;
	char *str;

	if ( v == NULL ) {
		val = 0;
		goto skip;
	}

	str = arg_get_string(v);
	if ( str == NULL )
		return -1;

#if _FILE_OFFSET_BITS == 64
	val = strtoull(str, &end, 0);
#else
	val = strtoul(str, &end, 0);
#endif
	if ( end == str )
		return -1;

	if ( val == ULONG_MAX && errno == ERANGE )
		return -1;

	switch( *end ) {
	case '\0':
		break;
	case 'k':
	case 'K':
		val <<= 10;
		break;
	case 'm':
	case 'M':
		val <<= 20;
		break;
	case 'g':
	case 'G':
		val <<= 30;
		break;
	case 't':
	case 'T':
		val <<= 40;
		break;
	case 'p':
	case 'P':
		val <<= 50;
		break;
	case 'e':
	case 'E':
		val <<= 60;
		break;
#if 0
	case 'z':
	case 'Z':
		val <<= 70;
		break;
	case 'y':
	case 'Y':
		val <<= 80;
		break;
#endif
	}

	if ( val > UINT_MAX )
		return -1;

skip:
	if ( a->type == ARGTYPE_POFF ) {
		if ( a->val.vp_off )
			*a->val.vp_off = val;
	}else{
		a->val.v_off = val;
	}

	return 1;
}

/* Unsigned integer data type */
static int arg_uint(struct arg *a, struct fvec *v, void *user)
{
	unsigned long int val;
	char *end;
	char *str;

	if ( v == NULL ) {
		val = 0;
		goto skip;
	}

	str = arg_get_string(v);
	if ( str == NULL )
		return -1;

	val = strtoul(str, &end, 0);
	if ( end == str )
		return -1;

	if ( *end != '\0' )
		return -1;

	if ( val == ULONG_MAX && errno == ERANGE )
		return -1;

	if ( val > UINT_MAX )
		return -1;

skip:
	if ( a->type == ARGTYPE_PUINT ) {
		if ( a->val.vp_uint )
			*a->val.vp_uint = val;
	}else{
		a->val.v_uint = val;
	}

	return 1;
}

/* Signed integer data type */
static int arg_int(struct arg *a, struct fvec *v, void *user)
{
	unsigned long int val;
	char *end, *str;
	int final = 0;

	if ( v == NULL ) {
		val = 0;
		goto skip;
	}

	str = arg_get_string(v);
	if ( str == NULL )
		return -1;

	/* negative */
	if ( *str == '-' ) {
		final = 1;
		str++;
	}

	val = strtoul(str, &end, 0);
	if ( end == str )
		return -1;

	if ( *end != '\0' )
		return -1;

	if ( val == ULONG_MAX && errno == ERANGE )
		return -1;

	if ( val > (UINT_MAX>>1) )
		return -1;

	if ( final )
		final = -val;
	else
		final = val;

skip:
	if ( a->type == ARGTYPE_PINT ) {
		if ( a->val.vp_int )
			*a->val.vp_int = val;
	}else{
		a->val.v_int = val;
	}

	return 1;
}

/* Double data type */
static int arg_double(struct arg *a, struct fvec *v, void *user)
{
	double val;
	char *end, *str;

	if ( v == NULL ) {
		val = 0.0f;
		goto skip;
	}

	str = arg_get_string(v);
	if ( str == NULL )
		return -1;

	val = strtod(str, &end);
	if ( end == str )
		return -1;

	if ( *end != '\0' )
		return -1;

	if ( val == HUGE_VAL && errno == ERANGE )
		return -1;

skip:
	if ( a->type == ARGTYPE_PDOUBLE ) {
		if ( a->val.vp_double )
			*a->val.vp_double = val;
	}else{
		a->val.v_double = val;
	}

	return 1;
}

/* String data type */
static int arg_string(struct arg *a, struct fvec *v, void *user)
{
	struct fvec def;

	if ( v == NULL ) {
		v = &def;
		v->ptr = NULL;
		v->len = 0;
	}

	if ( a->type == ARGTYPE_PSTRING ) {
		if ( a->val.vp_str )
			*a->val.vp_str = *v;
	}else{
		a->val.v_str = *v;
	}

	return 1;
}

static int arg_custom(struct arg *a, struct fvec *v, void *user)
{
	if ( a->val.v_custom == NULL )
		return -1;
	return a->val.v_custom(a, v, user);
}

/* Dispatch a key/value pair to the right place */
static int arg_found(struct arg *a,
	const char *key, size_t klen,
	const char *val, size_t vlen,
	void *user) {
	struct fvec kvec, vvec;
	struct arg *tmp;

	kvec.ptr = (char *)key;
	kvec.len = klen;
	vvec.ptr = (char *)val;
	vvec.len = vlen;

	for(tmp=a; tmp->name; tmp++) {
		if ( vstrcmp(&kvec, tmp->name) )
			continue;

		if ( tmp->type >= ARGTYPE_MAX || !arg_dfn[tmp->type] )
			return -1;

		if ( tmp->count && (tmp->type != ARGTYPE_CUSTOM) )
			return -1;

		tmp->count++;

		return arg_dfn[tmp->type](tmp, &vvec, user);
	}

	return -1;
}

int arg_dispatch(struct arg *a, const char *key, const char *val, void *priv)
{
	struct arg *tmp;

	for(tmp=a; tmp->name; tmp++)
		tmp->count = 0;

	return arg_found(a, key, strlen(key),
		val, val ? strlen(val) : 0, priv);
}

/* Determine what kind of quote character this is */
static inline int arg_quote(char q)
{
	int i=0;

	for(i=1; arg_q[i]; i++) {
		if ( arg_q[i] == q )
			return i;
	}

	i=0;
	return i;
}

/* Actual API entry */
int args_parse(struct arg *a, const char *str, void *user)
{
	struct arg *tmp;
	int ret=0;

	/* state machine tape position */
	const char *cur;

	/* Initial state */
	int state=0;
	int end=0;
	int inq=0;
	int esc=0;

	/* Key and value vectors */
	const char *key=NULL; size_t klen=0;
	const char *val=NULL; size_t vlen=0;

	if ( a == NULL )
		return -1;

	for(tmp=a; tmp->name; tmp++)
		tmp->count = 0;

	if ( str == NULL )
		goto do_defaults;

	for(cur=str; ; cur++) {
		if ( *cur==0 ) end=1;

		if ( state==0 ) { /* leading isspace */
			if ( !isspace(*cur) ) {
				state=1;
				key=cur;
				klen=1;
			}
		}else if ( state==1 ) { /* Inside a key */
			if ( *cur=='=' ) {
				val=cur+1;
				vlen=0;
				state=2;
			}else if ( isspace(*cur) || *cur==0 ) {
				if ( (ret=arg_found(a, key, klen,
					val, vlen, user))<=0 )
					return ret;
				state=0;
			}else klen++;
		}else if ( state==2 ) { /* Inside a value */
		
			/* Handle escaped characters */
			if ( esc > 2 ) esc=0;
			if ( !esc && *cur=='\\' ) {
				esc=1;
			}else if ( esc ){
				esc++;
			}

			if ( !inq ) {
				if ( !vlen && (inq=arg_quote(*cur)) ) {
					/* The first char is a quote char */
					val++;
				}else if ( isspace(*cur) || *cur=='\0') {
					/* No quotes */
					if ( !vlen )
						val=NULL;
					if ( (ret=arg_found(a, key, klen,
						val, vlen, user))<=0 )
						return ret;
					key=val=NULL;
					esc=state=inq=klen=vlen=0;
				}else vlen++;
			}else{
				/* Check for an end quote */
				if ( (!esc && *cur==arg_q[inq]) ||
					*cur==0) {
					if ( !vlen ) val=NULL;
					if ( (ret=arg_found(a, key, klen,
						val, vlen, user))<=0 )
						return ret;
					key=val=NULL;
					esc=state=inq=klen=vlen=0;
				}else vlen++;
			}
		}

		if ( end )
			break;
	}

	/* Fill in default values for any arguments that weren't used */
do_defaults:
	for(tmp=a; tmp->name; tmp++) {
		struct fvec v;

		if ( tmp->count )
			continue;

		/* Obliterate values of bad args */
		if ( tmp->type > ARGTYPE_MAX || !arg_dfn[tmp->type] ) {
			memset(&tmp->val, 0, sizeof(tmp->val));
			continue;
		}

		if ( tmp->def && *tmp->def != '\0' ) {
			v.ptr = (char *)tmp->def;
			v.len = strlen(tmp->def);
		}else{
			v.ptr = NULL;
			v.len = 0;
		}

		ret = (*arg_dfn[tmp->type])(tmp, v.len ? &v : NULL, user);
		if ( ret <= 0 )
			break;

#if 0
		mesg(M_DEBUG, "%s defaulted to %s",
			tmp->name,
			tmp->def ? tmp->def : "global default");
#endif
	}

	return ret;
}

const char *args_typename(unsigned int type)
{
	switch ( type ) {
	case ARGTYPE_CUSTOM:
		return "custom";
	case ARGTYPE_STRING:
	case ARGTYPE_PSTRING:
		return "string";
	case ARGTYPE_UINT:
	case ARGTYPE_PUINT:
		return "unsigned integer";
	case ARGTYPE_INT:
	case ARGTYPE_PINT:
		return "signed integer";
	case ARGTYPE_BYTES:
	case ARGTYPE_PBYTES:
		return "bytes";
	case ARGTYPE_BOOL:
	case ARGTYPE_PBOOL:
		return "bool";
	case ARGTYPE_DOUBLE:
	case ARGTYPE_PDOUBLE:
		return "double";
	case ARGTYPE_OFF:
	case ARGTYPE_POFF:
		switch ( sizeof(off_t) ) {
		case 4:
			return "32bit file offset";
		case 8:
			return "64bit file offset";
		default:
			return "file offset";
		}
	default:
		return "unknown";
	}
}
