/*
* This file is part of Firestorm NIDS
* Copyright (c) 2002,2003,2004 Gianni Tedesco
* This program is released under the terms of the GNU GPL version 2
*
* This is the decode engine. We keep a big internal graph of all
* protocols and all their dependancies on each other. The actual
* decoding is done by plugins.
*/
#include <firestorm.h>
#include <dotscara/fobuf.h>
#include <args.h>
#include <alert.h>
#include <signature.h>
#include <decode.h>
#include <loader.h>
#include <idx.h>
#include <i_idx.h>
#include <q_idx.h>
#include <q_filter.h>
#include <field.h>
#include <string_hash.h>

/* libpcap DLT_* protocol */
static struct proto pcap_dlt_p = init_proto("__pcap_dlt", NULL, NULL);
const struct proto *pcap_dlt = &pcap_dlt_p;

/* Global protocol list */
#define DHASH_SIZE 127
static struct proto *protos[DHASH_SIZE];
static unsigned int num_protos;

/* Internal protocol request list */
struct preq {
	struct proto		*proto;
	struct proto_req	*arr;
};
static struct mpool preq;

unsigned int decode_num_protos(void)
{
	return num_protos;
}

/* Find a protocol by name */
struct proto *decode_proto(const char *proto)
{
	struct proto *p;

	for(p=protos[string_hash(proto)%DHASH_SIZE]; p; p=p->next) {
		if ( !strcmp(p->name, proto) )
			return p;
	}

	return NULL;
}

static int ccmp(const void *A, const void *B)
{
	struct proto_child *a = (struct proto_child *)A;
	struct proto_child *b = (struct proto_child *)B;
	if ( a->id < b->id )
		return -1;
	if ( a->id > b->id )
		return 1;
	return 0;
}

/* Capture plugins call this   */
struct proto *decode_subproto(const char *proto, proto_id_t id)
{
	struct proto *p;
	struct proto_child k, *ret;

	p = decode_proto(proto);
	if ( p == NULL )
		return NULL;

	k.id = id;

	ret = bsearch(&k, p->child, p->num_child, sizeof(*p->child), ccmp);
	if ( ret )
		return ret->proto;

	return NULL;
}

struct proto *decode_next(struct proto *p, proto_id_t id)
{
	struct proto_child k, *ret;

	k.id = id;

	ret = bsearch(&k, p->child, p->num_child, sizeof(*p->child), ccmp);
	if ( ret )
		return ret->proto;

	return NULL;
}

/* plugins call this to register protocols */
static int decode_add(struct proto *p, struct proto_req *r)
{
	struct preq *req;
	u_int32_t h;

	if ( p == NULL || r == NULL )
		return 0;
	
	if ( p->name == NULL || p->decode == NULL )
		return 0;

	if ( strlen(p->name) > PROTO_LEN )
		mesg(M_WARN, "decode: %s: protocol name is too long", p->name);

	/* Check if we have already been added */
	if ( p->next || p->num_child || p->child || decode_proto(p->name) ) {
		mesg(M_DEBUG,"decode: error: '%s' protocol is "
			"already registered", p->name);
		return 0;
	}

	/* XXX: Should we make all fields field_reg's instead? */
	if ( p->fields ) {
		struct field *ptr = p->fields;

		p->fields = NULL;
		field_add(p, ptr);
	}

	/* Save the requests for later, when all the protocols are done
	 * registering. We resolve them in decode_resolve_all()
	 */
	req = mpool_alloc(&preq);
	if ( req == NULL )
		firestorm_die_error("calloc");

	req->proto = p;
	req->arr = r;

	/* Add to the hash table */
	h = string_hash(p->name) % DHASH_SIZE;
	p->next = protos[h];
	protos[h] = p;
	p->idx = num_protos++;

	return 1;
}

/* Build the graph of inter-dependencies for all registered protocols */
static int decode_resolve_all(void)
{
	struct mpool_hdr *h, *f;
	struct proto_req *reqs;
	struct proto *child, *p;
	void *ptr;
	int i;

	for(h=preq.slabs; (f=h); h=h->next) {
		for(ptr=f->data; ptr < f->next_obj; ptr += preq.obj_size) {
			reqs = ((struct preq *)ptr)->arr;
			child = ((struct preq *)ptr)->proto;

			for(; reqs->name; reqs++) {
				p = decode_proto(reqs->name);
				if ( p )
					p->num_child++;
			}
		}
	}

	for(h=preq.slabs; (f=h); h=h->next) {
		for(ptr=f->data; ptr < f->next_obj; ptr += preq.obj_size) {
			reqs = ((struct preq *)ptr)->arr;
			child = ((struct preq *)ptr)->proto;

			for(; reqs->name; reqs++) {
				p = decode_proto(reqs->name);
				if ( p == NULL )
					continue;

				if ( p->child == NULL ) {
					p->child = calloc(p->num_child,
							sizeof(*p->child));
					p->num_child = 0;
				}

				if ( p->child == NULL )
					goto err;

				p->child[p->num_child].id = reqs->id;
				p->child[p->num_child].proto = child;
				p->num_child++;
			}
		}
	}

	mpool_free(&preq);

	for(i=0; i < DHASH_SIZE; i++)
		for(p=protos[i]; p; p=p->next)
			qsort(p->child, p->num_child, sizeof(*p->child), ccmp);

	return 1;

err:
	mesg(M_ERR, "decode: resolving protocols: %s", sys_err2(NULL));
	mpool_free(&preq);
	return 0;
}

static void decode_register(struct plugin *p, struct proto_table *t)
{
	for(; t->proto; t++)
		decode_add(t->proto, t->preq);
}

struct field *field_find(const char *name)
{
	struct field *f;
	struct proto *p;
	int i;

	for(f=field_builtins(); f; f=f->next) {
		if ( !strcmp(name, f->name) )
			return f;
	}

	for(i=0; i < DHASH_SIZE; i++) {
		for(p=protos[i]; p; p=p->next)
			for(f=p->fields; f; f=f->next)
				if ( !strcmp(name, f->name) )
					return f;
	}

	return NULL;
}

struct field *field_foreach(field_cbfn_t cbfn, void *priv)
{
	struct field *f;
	struct proto *p;
	int i;

	for(f=field_builtins(); f; f=f->next)
		if ( !(*cbfn)(f, priv) )
			return f;

	for(i=0; i < DHASH_SIZE; i++) {
		for(p=protos[i]; p; p=p->next)
			for(f=p->fields; f; f=f->next)
				if ( !(*cbfn)(f, priv) )
					return f;
	}

	return NULL;
}

void decode_load(void)
{
	loader_find_syms("firestorm_protocols", (loader_cb_t)decode_register);
	decode_resolve_all();
}

/* Constructor manually adds __pcap_dlt to the table */
void __private decode_init(void)
{
	u_int32_t h = string_hash((char *)pcap_dlt_p.name)%DHASH_SIZE;
	protos[h] = &pcap_dlt_p;
	mpool_new(&preq, sizeof(struct preq), 0);
}
