/*
* This file is part of dotscara
* Copyright (c) 2003 Gianni Tedesco
* Released under the terms of the GNU GPL version 2
*/

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <limits.h>
#include <sys/types.h>

#include <dotscara/dotscara.h>
#include <dotscara/initmem.h>

/** initmem_new
 * @param m: An initmem structure to use
 * @param slab_size: default size of allocation units
 *
 * Creates a new initmem allocator descriptor with the passed values
 * set.
 *
 * Return value: zero on error, non-zero for success.
 */
int __public initmem_new(struct initmem *m, size_t slab_size)
{
	if ( slab_size == 0 )
		slab_size = 8192;
	if ( slab_size < sizeof(struct initmem_slab) )
		slab_size = sizeof(struct initmem_slab);

	m->slab_size = slab_size;
	m->head = NULL;
	m->tail = NULL;

	return 1;
}

/** initmem_free
 * @param m: The initmem structure to free up
 *
 * Frees all memory allocated by the passed initmem descriptor and
 * resets all members to invalid values.
 */
void __public initmem_free(struct initmem *m)
{
	struct initmem_slab *s;

	for(s=m->head; s;) {
		struct initmem_slab *t = s;
		s = s->next;
		free(t);
	}

	memset(m, 0, sizeof(*m));
}

/** __initmem_alloc_slow
 * @param m: The initmem structure to allocate from
 * @param sz: Object size
 *
 * Slow path for allocations which may call allocate more system
 * memory.
*/
void __public *__initmem_alloc_slow(struct initmem *m, size_t sz)
{
	size_t alloc;
	size_t overhead;
	struct initmem_slab *s;
	void *ret;

	overhead = sizeof(struct initmem_slab);

	alloc = overhead + sz;
	if ( alloc < m->slab_size )
		alloc = m->slab_size;

	s = malloc(alloc);
	if ( s == NULL )
		return NULL;

	ret = (void *)s + overhead;
	s->ptr = (void *)s + overhead + sz;
	s->len = alloc - (overhead + sz);
	s->next = NULL;

	if ( m->tail == NULL ) {
		m->head = s;
	}else{
		m->tail->next = s;
	}

	m->tail = s;

	return ret;
}
