/*
 * This file is part of Firestorm NIDS
 * Copyright (c) 2004 Gianni Tedesco
 * Released under the terms of the GNU GPL version 2
 *
 * TODO:
 *  o Advanced error handling
 *  o Integrate mesg()
 *  o Integrate querying/signatures
 *  o Arbitrary pipeline construction
*/
#include <firestorm.h>
#include <signal.h>
#include <unistd.h>
#include <capdev.h>
#include <exporter.h>
#include <alert.h>
#include <signature.h>
#include <decode.h>
#include <preproc.h>

static struct intr_ops *intr_ops;
static void *intr_priv;
static volatile sig_atomic_t intr_pending;

struct per_proto {
	struct list_head preproc;
};

/* ------ PRIVATE API ----- */
static void do_source_del(struct source *s)
{
	list_move(&s->list, &s->owner->d_sources);
}

static void pump_deleted_sources(struct pipeline *p)
{
	struct source *s, *tmp;

	list_for_each_entry_safe(s, tmp, &p->d_sources, list) {
		mesg(M_INFO , "source: %s[%s]: stopped",
			s->plugin->name, s->args);
		list_del(&s->list);
		s->plugin->free(s);
		free(s);
	}
}

static void do_source_del_all_and_pump(struct pipeline *p)
{
	list_splice_init(&p->f_sources, &p->d_sources);
	list_splice_init(&p->l_sources, &p->d_sources);
	pump_deleted_sources(p);
}

static int do_sink_close(struct sink *s)
{
	int ret;
	list_del(&s->list);
	ret = s->plugin->free(s);
	free(s);
	return ret;
}

static int do_sink_close_all(struct pipeline *p)
{
	int h;
	int ret = 1;

	for(h=0; h < SINK_MAX; h++) {
		struct sink *s, *tmp;

		list_for_each_entry_safe(s, tmp, &p->ohook[h], list)
			if ( !do_sink_close(s) )
				ret = 0;
	}

	return ret;
}

static int do_sink(struct pipeline *p, int h, struct packet *pkt)
{
	struct sink *s, *tmp;
	int ret = 1;

	list_for_each_entry_safe(s, tmp, &p->ohook[h], list) {
		if ( !s->plugin->queue(s, pkt) ) {
			ret = 0;
			if ( (p->flags & PIPELINE_SINK_ERRORS_FATAL) )
				return ret;
		}
	}

	return ret;
}

struct mergectx {
	struct source **pending;
	unsigned int num_pending;
};

static int cm_cmp(const void *A, const void *B)
{
	const struct source *a = *(const struct source **)A;
	const struct source *b = *(const struct source **)B;
	if ( a->last_pkt == NULL )
		return 1;
	if ( b->last_pkt == NULL )
		return -1;
	if ( time_before(a->last_pkt->time, b->last_pkt->time) )
		return -1;
	if ( time_after(a->last_pkt->time, b->last_pkt->time) )
		return 1;
	return 0;
}

static struct packet *capdev_dequeue(struct source *s)
{
	s->last_pkt = s->plugin->dequeue(s);
	return s->last_pkt;
}

static int mergectx_init(struct mergectx *ctx, struct pipeline *p)
{
	struct source *s, *tmp;
	struct source **pending;
	unsigned int i = 0;

	if ( p->num_f_sources == 0 )
		return 1;

	ctx->pending = NULL;
	ctx->num_pending = 0;

	pending = malloc(p->num_f_sources * sizeof(*pending));
	if ( pending == NULL )
		return 0;

	list_for_each_entry_safe(s, tmp, &p->f_sources, list)
		if ( capdev_dequeue(s) )
			pending[i++] = s;

	ctx->pending = pending;
	pending = realloc(pending, i * sizeof(*pending));
	if ( pending == NULL ) {
		free(ctx->pending);
		ctx->pending = NULL;
		return 0;
	}

	qsort(pending, i, sizeof(*pending), cm_cmp);

	ctx->pending = pending;
	ctx->num_pending = i;

	pump_deleted_sources(p);
	return 1;
}

static struct packet *mergectx_first_pkt(struct mergectx *ctx)
{
	if ( ctx->num_pending )
		return ctx->pending[0]->last_pkt;
	return NULL;
}

static struct packet *mergectx_next_pkt(struct mergectx *ctx)
{
	if ( ctx->num_pending == 0 )
		return NULL;

	if ( capdev_dequeue(ctx->pending[0]) == NULL ) {
		ctx->num_pending--;
		do_source_del(ctx->pending[0]);
		memmove(&ctx->pending[1], &ctx->pending[0],
			ctx->num_pending * sizeof(*ctx->pending));
	}

	if ( ctx->num_pending ) {
		qsort(ctx->pending, ctx->num_pending,
			sizeof(*ctx->pending), cm_cmp);
		return ctx->pending[0]->last_pkt;
	}

	return NULL;
}

static void mergectx_free(struct mergectx *ctx)
{
	free(ctx->pending);
	ctx->pending = NULL;
	ctx->num_pending = 0;
}

static int handle_intrs(void)
{
	sig_atomic_t intrs;

	if ( intr_ops )
		intr_ops->lock(intr_priv);

	intrs = intr_pending;
	intr_pending = 0;

	if ( intr_ops )
		intr_ops->unlock(intr_priv);

	if ( intrs & PIPELINE_INTR_HUP )
		notify_run(NOTIFY_HUP, NULL, -1);
	if ( intrs & PIPELINE_INTR_ITIMER )
		notify_run(NOTIFY_ITIMER, NULL, -1);
	if ( intrs & PIPELINE_INTR_VTIMER )
		notify_run(NOTIFY_ITIMER, NULL, -1);
	if ( intrs & PIPELINE_INTR_USR1 )
		notify_run(NOTIFY_USR1, NULL, -1);
	if ( intrs & PIPELINE_INTR_USR2 )
		notify_run(NOTIFY_USR2, NULL, -1);

	if ( intrs & PIPELINE_INTR_DEADLY )
		return 0;

	return 1;
}

/* return zero for OK, -1 to suppress */
static int alert_token_bucket(struct tokenbucket *tb,
				struct packet *p,
				const char *s1, const char *s2)
{
	int missed;

	if ( !tb )
		return 1;

	switch ( (missed=token_bucket(tb, p->time)) ) {
	case -1:
		return 0;
	case 0:
		return 1;
	default:
		mesg(M_DEBUG, "alert: [%s] %s: %i alerts suppressed",
			s1, s2, missed);
		return 1;
	}
}

static int inject_packet(struct packet *p)
{
	if ( p->flags & FP_REAL )
		p->source->owner->tot_bytes += p->caplen;

	if ( p->source->plugin->flags & CAPDEV_REALTIME )
		time_update_now(p->time);

	if ( intr_pending && !handle_intrs() )
		return 0;

	if (  p->layer[p->llen].proto ||
		p->llen >= PKT_LAYERS ) {
		p->layer[p->llen].proto->decode(p);
	}else{
		dispatch(p);
	}

	/* return -1 for some fatal error */
	if ( p->source->owner->status == PIPELINE_STATUS_STOPPED )
		return -1;

	return 1;
}

static int file_seq_mainloop(struct pipeline *p)
{
	struct source *s, *tmp;

	list_for_each_entry_safe(s, tmp, &p->f_sources, list) {
		struct packet *pkt;

		while ( (pkt = capdev_dequeue(s)) ) {
			switch ( inject_packet(pkt) ) {
			case -1:
				return 0;
			case 0:
				return 1;
			default:
				continue;
			}
		}

		do_source_del(s);
		pump_deleted_sources(p);
	}

	return 1;
}

static int file_merge_mainloop(struct pipeline *p)
{
	struct mergectx ctx;
	struct packet *pkt;

	if ( !mergectx_init(&ctx, p) )
		return 0;

	pkt = mergectx_first_pkt(&ctx);
again:
	if ( pkt == NULL )
		return 1;

	switch( inject_packet(pkt) ) {
	case -1:
		mergectx_free(&ctx);
		return 0;
	case 0:
		mergectx_free(&ctx);
		return 1;
	default:
		pkt = mergectx_next_pkt(&ctx);
		pump_deleted_sources(p);
		goto again;
	}
}

static int nbio_mainloop(struct pipeline *p)
{
	struct mergectx ctx;
	struct packet *pkt;
	struct nbio *n, *tmp;

	if ( !mergectx_init(&ctx, p) )
		return 0;

	pkt = mergectx_first_pkt(&ctx);
again:

	list_for_each_entry_safe(n, tmp, &p->iothread.active, list) {
		struct source *s = n->priv;
		struct packet *lpkt;

		lpkt = capdev_dequeue(s);
		if ( lpkt == NULL )
			continue;

		while ( pkt &&
			time_before(pkt->time, lpkt->time) ) {
			switch ( inject_packet(pkt) ) {
			case -1:
				goto err;
			case 0:
				goto intr;
			default:
				pkt = mergectx_next_pkt(&ctx);
			}
		}

		switch ( inject_packet(lpkt) ) {
		case -1:
			goto err;
		case 0:
			goto intr;
		default:
			break;
		}
	}

	if ( !list_empty(&p->iothread.active) )
		goto again;

	pump_deleted_sources(p);

	nbio_pump(&p->iothread);

	if ( intr_pending && !handle_intrs() )
		return 1;

	goto again;
err:
	mergectx_free(&ctx);
	return 0;
intr:
	mergectx_free(&ctx);
	return 1;
}

static struct list_head *preproc_cb(const char *protocol, void *priv)
{
	struct pipeline *p = priv;
	struct proto *proto;

	if ( protocol == NULL )
		return &p->hook[HOOK_PRE_PROCESSING];

	proto = decode_proto(protocol);
	if ( proto == NULL )
		return NULL;

	return &p->proto[proto->idx].preproc;
}

/* ------ PUBLIC API ----- */
void pipeline_intr_ops(struct intr_ops *ops, void *priv)
{
	intr_ops = ops;
	intr_priv = priv;
}

void pipeline_intr(bitmask_t intrs, int reentrant)
{
	if ( reentrant && intr_ops )
		intr_ops->lock(intr_priv);

	intr_pending |= intrs;

	if ( reentrant && intr_ops )
		intr_ops->unlock(intr_priv);
}

int pipeline_init(struct pipeline *p, bitmask_t flags)
{
	unsigned int num_proto = decode_num_protos();
	unsigned int i;

	p->proto = malloc(num_proto * sizeof(*p->proto));
	if ( num_proto && (p->proto == NULL) )
		return 0;

	for(i=0; i < num_proto; i++) {
		INIT_LIST_HEAD(&p->proto[i].preproc);
	}

	if ( (flags & PIPELINE_ALLOW_LIVE) &&
		!nbio_init(&p->iothread, NULL) ) {
		free(p->proto);
		return 0;
	}

	p->flags = flags;
	p->status = PIPELINE_STATUS_INIT;

	INIT_LIST_HEAD(&p->f_sources);
	INIT_LIST_HEAD(&p->l_sources);
	INIT_LIST_HEAD(&p->d_sources);
	INIT_LIST_HEAD(&p->sinks);

	p->num_f_sources = 0;
	p->num_l_sources = 0;

	for(i=0; i < HOOK_MAX; i++)
		INIT_LIST_HEAD(&p->hook[i]);
	for(i=0; i < SINK_MAX; i++)
		INIT_LIST_HEAD(&p->ohook[i]);
	return 1;
}

void pipeline_free(struct pipeline *p)
{
	int h;

	p->status = PIPELINE_STATUS_STOPPED;

	/* Kill off nbio captures */
	if ( p->flags & PIPELINE_ALLOW_LIVE )
		nbio_fini(&p->iothread);

	/* Delete any remaining sources */
	do_source_del_all_and_pump(p);

	/* Delete all hooks */
	for(h=0; h < HOOK_MAX; h++)
		notifier_delete(&p->hook[h]);

	/* Delete all sinks */
	do_sink_close_all(p);

	/* Remove per-protocol structures */
	if ( p->proto )
		free(p->proto);

	p->flags = 0;
}

struct source *pipeline_source_add(struct pipeline *p,
					const char *plugin,
					const char *args)
{
	struct source *ret = NULL;
	struct capdev *cd;
	size_t len;

	cd = capdev_find(plugin);
	if ( cd == NULL )
		goto err;

	if ( (cd->flags & CAPDEV_NBIO) &&
		!(p->flags & PIPELINE_ALLOW_LIVE) )
		goto err;

	if ( !(cd->flags & CAPDEV_NBIO) &&
		!(p->flags & PIPELINE_ALLOW_FILE) )
		goto err;

	len = (args) ? (strlen(args) + 1) : (0);
	ret = malloc(sizeof(*ret) + len);
	if ( ret == NULL )
		goto err;

	ret->plugin = cd;
	ret->last_pkt = NULL;
	ret->owner = p;
	memcpy(ret->args, args, len);

	if ( !ret->plugin->init(ret, args) )
		goto err;

	if ( cd->flags & CAPDEV_NBIO ) {
		list_add_tail(&ret->list, &p->l_sources);
		p->num_l_sources++;
	}else{
		list_add_tail(&ret->list, &p->f_sources);
		p->num_f_sources++;
	}

	mesg(M_INFO , "source: %s[%s]: started",
		ret->plugin->name, ret->args);
	return ret;

err:
	if ( ret )
		free(ret);
	return NULL;
}

int pipeline_is_live(struct pipeline *p)
{
	return !!p->num_l_sources;
}

int pipeline_preproc_add(struct pipeline *p,
				const char *plugin,
				const char *args)
{
	struct preproc *preproc;

	preproc = preproc_find(plugin);
	if ( preproc == NULL )
		return 0;

	return preproc->init(args, preproc_cb, p);
}

struct sink *pipeline_sink_add(struct pipeline *p,
				int h,
				const char *plugin,
				const char *args)
{
	struct sink *s;
	struct exporter *e;
	size_t len = 0;

	if ( (h < 0) || (h >= SINK_MAX) )
		return NULL;

	e = exporter_find(plugin);
	if ( e == NULL )
		return NULL;

	len = (args) ? (strlen(args) + 1) : (0);
	s = malloc(sizeof(*s) + len);
	if ( s == NULL )
		return 0;

	s->plugin = e;
	s->priv = NULL;
	s->owner = p;
	memcpy(s->args, args, len);

	if ( !s->plugin->init(s, args) ) {
		free(s);
		return NULL;
	}

	list_add_tail(&s->list, &p->ohook[h]);

	return s;
}

int pipeline_sink_close(struct sink *s)
{
	return do_sink_close(s);
}

int pipeline_sink_close_all(struct pipeline *p)
{
	return do_sink_close_all(p);
}

void pipeline_rewind(struct pipeline *p)
{
	struct source *s, *tmp;

	list_for_each_entry_safe(s, tmp, &p->f_sources, list) {
		if ( !s->plugin->rewind ) {
			mesg(M_WARN, "%s: no rewind method",
					s->plugin->name);
			continue;
		}
		s->plugin->rewind(s);
	}
}

int pipeline_go(struct pipeline *p)
{
	timestamp_t tm;
	timestamp_t vtm;
	u_int64_t tmp;
	int ret;
	int (*ml)(struct pipeline *p);

	if ( list_empty(&p->l_sources) &&
		list_empty(&p->f_sources) )
		return 0;
	
	p->status = PIPELINE_STATUS_RUNNING;
	p->tot_bytes = 0;

	if ( list_empty(&p->l_sources) ) {
		if ( (p->flags & PIPELINE_MERGE) &&
			(p->num_f_sources > 1) ) {
			ml = file_merge_mainloop;
		}else{
			ml = file_seq_mainloop;
		}
	}else{
		if ( !(p->flags & PIPELINE_MERGE) &&
			p->num_f_sources ) {
			mesg(M_WARN, "merge flag implicitly enabled by "
				"selecting both file and live captures");
			p->flags |= PIPELINE_MERGE;
		}
		ml = nbio_mainloop;
	}

	/* XXX: UNIX specific */
	if ( p->flags & PIPELINE_PROFILE_MODE )
		pause();

	tm = time_gettime();
	vtm = time_getvtime();

	ret = ml(p);

	tm = time_gettime() - tm;
	vtm = time_getvtime() - vtm;

	/* XXX: UNIX specific */
	if ( p->flags & PIPELINE_PROFILE_MODE )
		pause();

	mesg(M_INFO, "elapsed time: %llu.%.9u seconds",
		time_secs(tm), time_nsec(tm));
	mesg(M_INFO, "cpu time: %llu.%.9u seconds",
		time_secs(vtm), time_nsec(vtm));

	tmp = (p->tot_bytes * 8 * TIMESTAMP_HZ) / tm;
	mesg(M_INFO, "avg. throughput: %llu.%.6llu mbps",
			tmp / (1000000), tmp % (1000000));

	tmp = (p->tot_bytes * 8 * TIMESTAMP_HZ) / vtm;
	mesg(M_INFO, "processing rate: %llu.%.6llu mbps",
			tmp / (1000000), tmp % (1000000));

	p->status = PIPELINE_STATUS_STOPPED;

	return ret;
}

struct notifier *pipeline_hook_add(struct pipeline *p, int h,
					notifier_t n,
					notifier_dtor_t d,
					void *priv)
{
	if ( (h < 0) || (h >= HOOK_MAX) )
		return NULL;

	return notifier_add(&p->hook[h], n, d, priv);
}

void pipeline_hook_delete(struct pipeline *p, int h)
{
	if ( (h < 0) || (h >= HOOK_MAX) )
		return;

	notifier_delete(&p->hook[h]);
}

void pipeline_hook_delete_all(struct pipeline *p)
{
	int h;

	for(h=0; h < HOOK_MAX; h++)
		notifier_delete(&p->hook[h]);
}

void dispatch(struct packet *pkt)
{
	struct pipeline *p = pkt->source->owner;
	struct layer *l;
	struct alert *a;
	int i, passed = 1;
	int ret = 1;

	/* Preprocessors and processors may set alerts */
	for(i=0, l=pkt->layer; i < pkt->llen; i++,l++) {
		struct list_head *preproc;

		if ( l->proto == NULL )
			continue;

		preproc = &p->proto[l->proto->idx].preproc;
		notifier_run(preproc, pkt, i);
	}
	notifier_run(&p->hook[HOOK_PRE_PROCESSING], pkt, -1);
	notifier_run(&p->hook[HOOK_PROCESSING], pkt, -1);

	/* So see if they set an alert, and see if it passes the token
	 * bucket filters
	 */
	a = pkt->alert;
	if ( a && !alert_token_bucket(&a->t, pkt, "sig", a->alert) &&
		!alert_token_bucket(a->gen->t, pkt, "generator", a->gen->name) )
		passed = 0;

	/* Perform alert hooks */
	if ( a ) {
		notifier_run(&p->hook[HOOK_PRE_ALERT], pkt, -1);
		if ( passed )
			notifier_run(&p->hook[HOOK_ALERT], pkt, -1);
	}

	/* Send packet to sinks */
	if ( !do_sink(p, SINK_ALL, pkt) )
		ret = 0;
	if ( pkt->flags & FP_REAL )
		if ( !do_sink(p, SINK_REAL, pkt) )
			ret = 0;

	if ( a ) {
		if ( !do_sink(p, SINK_PRE_ALERT, pkt) )
			ret = 0;
		if ( passed )
			if ( !do_sink(p, SINK_ALERT, pkt) )
				ret = 0;
	}

	if ( ret )
		return;

	if ( p->flags & PIPELINE_SINK_ERRORS_FATAL ) {
		p->status = PIPELINE_STATUS_STOPPED;
	}else{
		mesg(M_CRIT, "fatal pipeline failure");
	}
}
